package autoweka;

import weka.core.Utils;
import weka.core.converters.ConverterUtils.DataSource;
import weka.core.Instances;
import weka.classifiers.Evaluation;
import weka.classifiers.Classifier;
import weka.attributeSelection.ASSearch;
import weka.attributeSelection.ASEvaluation;
import java.io.File;
import java.io.FileFilter;

import java.util.ArrayList;
import java.util.List;

/**
 * Utility class for determining what kind of methods can be applied to a set of instances.
 *
 * Goes through a folder looking for .param files, and trys to instantiate the class to see if 
 * it can be used with the given set of instances 
 */
class ApplicabilityTester
{
    //What are we searching for
    private enum Testable{
        CLASSIFIER,
        ATTRIBUTE_SEARCH,
        ATTRIBUTE_EVAL
    };
    
    public static ArrayList<String> getApplicableClassifiers(Instances instances, String paramDirName)
    {
        return getApplicable(instances, paramDirName, Testable.CLASSIFIER, null);
    }

    /**
     * Looks inside a folder for .param files that are WEKA classifiers
     *
     * @param instances The instances to test on
     * @param paramDirName The path to the folder containing the .param files
     */
    public static ArrayList<String> getApplicableClassifiers(Instances instances, String paramDirName, List<String> allowed)
    {
        return getApplicable(instances, paramDirName, Testable.CLASSIFIER, allowed);
    }
    /**
     * Looks inside a folder for .param files that are WEKA ASEvals
     *
     * @param instances The instances to test on
     * @param paramDirName The path to the folder containing the .param files
     */
    public static ArrayList<String> getApplicableAttributeEvaluators(Instances instances, String paramDirName)
    {
        return getApplicable(instances, paramDirName, Testable.ATTRIBUTE_EVAL, null);
    }
    
    /**
     * Looks inside a folder for .param files that are WEKA ASSearches
     *
     * @param instances The instances to test on
     * @param paramDirName The path to the folder containing the .param files
     */
    public static ArrayList<String> getApplicableAttributeSearchers(Instances instances, String paramDirName)
    {
        return getApplicable(instances, paramDirName, Testable.ATTRIBUTE_SEARCH, null);
    }

    private static ArrayList<String> getApplicable(Instances instances, String paramDirName, Testable type, List<String> allowed)
    {
        File paramDir = new File(paramDirName);
        if(!paramDir.exists() || !paramDir.isDirectory())
        {
            throw new RuntimeException(paramDirName + " is not a valid directory");
        }
        FileFilter paramFilter = new FileFilter() {
            public boolean accept(File file) {
                //Go find param files, and ignore stuff that vim leaves behind when I'm editing it..
                return file.getName().endsWith(".params") && !file.getName().startsWith(".");
            }
        };

        ArrayList<String> names = new ArrayList<String>();
        File[] params = paramDir.listFiles(paramFilter);
        for(File i:params)
        {
           names.add(Util.removeExtension(i.getName()));
        }
        java.util.Collections.sort(names);

        ArrayList<String> goodMethods = new ArrayList<String>();

        for(String name: names)
        {
            if(allowed != null && !allowed.contains(name))
                continue;
            if(isApplicable(name, instances, type))
            {
                goodMethods.add(name);
            }
        }
        return goodMethods;
    }

    private static boolean isApplicable(String method, Instances instances, Testable type)
    {
        switch(type){
            case CLASSIFIER:
                return isApplicableClassifier(method, instances);
            case ATTRIBUTE_EVAL:
                return isApplicableAttributeEvaluator(method, instances);
            case ATTRIBUTE_SEARCH:
                return isApplicableAttributeSearch(method, instances);
            default:
                return false;
        }
    }
    
    private static boolean isApplicableClassifier(String method, Instances instances)
    {
        //Go get ourselves a classifier, and get some 
        try
        {
            Class<?> cls = Class.forName(method);
            Classifier classifier = (Classifier)cls.newInstance();
            classifier.getCapabilities().testWithFail(instances);
            return true;
        }
        catch(ClassNotFoundException e)
        {
            System.out.println("No class '" + method + "' found");
        }
        catch(InstantiationException e)
        {
            System.out.println("Failed to instantiate '" + method + "': " + e.getMessage());
        }
        catch(IllegalAccessException e)
        {
            System.out.println("Illegal access exception creating '" + method + "': " + e.getMessage());
        }
        catch(weka.core.UnsupportedAttributeTypeException e)
        {
            System.out.println(method + " failed with message: " + e.getMessage());
        }
        catch(Exception e)
        {
            e.printStackTrace();
            System.out.println("'" + method + "' not supported");
        }
        return false;
    }

    private static boolean isApplicableAttributeSearch(String method, Instances instances){
        //Build an AS...Search  (curse you bad naming scheme!) and see if it can handle the instances
        try{
            ASSearch search = ASSearch.forName(method, new String[0]);
        } catch(weka.core.UnsupportedAttributeTypeException e) {
            System.out.println(method + " failed with message: " + e.getMessage());
        }catch(Exception e){
            e.printStackTrace();
            System.out.println("'" + method + "' not supported");
            return false;
        }
        return true;
    }
    
    private static boolean isApplicableAttributeEvaluator(String method, Instances instances){
        //Build an ASE and see if it can handle the instances
        try{
            ASEvaluation eval = ASEvaluation.forName(method, new String[0]);
            eval.getCapabilities().testWithFail(instances);
        } catch(weka.core.UnsupportedAttributeTypeException e) {
            System.out.println(method + " failed with message: " + e.getMessage());
        }catch(Exception e){
            e.printStackTrace();
            System.out.println("'" + method + "' not supported");
            return false;
        }
        return true;
    }
}
