package autoweka;

import weka.core.Instances;
import java.io.File;
import java.util.Arrays;
import java.util.Collections;
import java.util.LinkedList;
import java.util.List;
import java.util.Map;
import java.util.Properties;
import java.util.Queue;
import java.util.HashMap;

import java.util.ArrayList;

/** 
 * Abstract class responsible for generating all the necessary stuff to run an Auto-WEKA Experiment.
 *
 * Although this class is static, you should use it's main method to actually do the experiment construction.
 * In particular, if you run it with cmd line arguments of a number of ExperimentBatch XML files, it will do
 * the appropriate generation for you by calling the right classes
 */
public abstract class ExperimentConstructor
{
    /**
     * The directory with all the param files in it. 
     */
    protected String mParamBaseDir = "./params";

    protected List<ClassParams> mBaseClassParams = new ArrayList<ClassParams>();
    protected List<ClassParams> mMetaClassParams = new ArrayList<ClassParams>();
    protected List<ClassParams> mEnsembleClassParams = new ArrayList<ClassParams>();
    protected List<ClassParams> mAttribSearchClassParams = new ArrayList<ClassParams>();
    protected List<ClassParams> mAttribEvalClassParams = new ArrayList<ClassParams>();
    /**
     * List containing what classifiers are allowed to be used by this experiment
     */
    protected List<String> mAllowedClassifiers = new ArrayList<String>();
    /**
     * How deep should the ensemble tree go?
     */
    protected int mEnsembleMaxNum = 5;
    //These flags sort of have meaning....
    protected boolean mIncludeBase = true;
    protected boolean mIncludeMeta = true;
    protected boolean mIncludeEnsemble = true;

    /**
     * The output path for the experiments
     */
    protected String mExperimentPath = "./experiments/";

    /**
     * The active Experiment that we're trying to build
     */
    protected Experiment mExperiment = null;

    /**
     * Properties associated with our constructor.
     *
     * These are defined in a file in the CWD that has the form CANONICAL_CLASS_NAME.properties
     */
    protected Properties mProperties = null;

    /**
     * The instance generator we're using, all loaded up and ready to go with the appropriate dataset 
     */
    protected InstanceGenerator mInstanceGenerator = null;


    /**
     * Main method that can either take stuff on the command line to build a single experiment, or points to an ExperimentBatch XML file.
     *
     * You can specify all the options that are in an XML file by using the same XML tag name preceeded with two dashes
     */
    public static void main(String[] args)
    {
        //Is the first argument a -batch? If it is, then we need to load the given xml files and use those to generate things
        if(args[0].equals("-batch") || new File(args[0]).isFile())
        {
            for(int i = 0; i < args.length; i++)
            {
                if(!args[i].startsWith("-"))
                    generateBatches(args[i]);
            }
        }
        else
        {
            LinkedList<String> argList = new LinkedList<String>(Arrays.asList(args));
            String constructorName = argList.poll();
            Experiment exp = new Experiment();
            XmlSerializable.populateObjectFromCMDParams(exp, argList);
            buildSingle(constructorName, exp, argList);
        }
    }

    private static void generateBatches(String xmlFile)
    {
        try
        {
            ExperimentBatch batch = ExperimentBatch.fromXML(xmlFile);


            //For each type of experiment component
            for(ExperimentBatch.ExperimentComponent expComp: batch.mExperiments)
            {
                //For each dataset component
                for(ExperimentBatch.DatasetComponent datasetComp: batch.mDatasets)
                {
                    Experiment exp = ExperimentBatch.createExperiment(expComp, datasetComp);

                    //Use the extra args from the expComp to give the runner
                    buildSingle(expComp.constructor, exp, new LinkedList<String>(expComp.constructorArgs));
                }
            }
        }
        catch(Exception e)
        {
            e.printStackTrace();
            throw new RuntimeException("Failed to create batch for " + xmlFile);
        }
    }

    private static void buildSingle(String builderClassName, Experiment exp, List<String> args)
    {
        exp.validate();

        //The first parameter contains the full class of the experiment constructor
        System.out.println("Making Experiment " + exp.name);
        Class<?> cls;
        ExperimentConstructor builder;
        try
        {
            cls = Class.forName(builderClassName);
            builder = (ExperimentConstructor)cls.newInstance();
        }
        catch(ClassNotFoundException e)
        {
            throw new RuntimeException("Could not find class '" + builderClassName + "': " + e.getMessage());
        }
        catch(Exception e)
        {
            //e.getCause().printStackTrace();
            throw new RuntimeException("Failed to instantiate '" + builderClassName + "': " + e.getMessage());
        }

        builder.run(exp, new LinkedList<String>(args));
    }

    private void run(Experiment exp, List<String> args)
    {
        //See if we can load up this constructor's canonical class name's props file
        mProperties = new Properties();
        try
        {
            mProperties.load(new java.io.FileInputStream(new java.io.File(this.getClass().getCanonicalName() + ".properties")));
        }
        catch(java.io.FileNotFoundException e)
        {
            System.out.println("No property file '" + this.getClass().getCanonicalName() + ".properties' found");
        }
        catch(java.io.IOException e)
        {
            e.printStackTrace();
        }

        if(exp.resultMetric == null){
            throw new RuntimeException("No Result Metric defined");
        }

        mExperiment = exp;
        Queue<String> argQueue = new LinkedList<String>(args);
        while(!argQueue.isEmpty())
        {
            String arg = argQueue.poll();
            if(arg.equals("-nometa"))
                mIncludeMeta = false;
            else if (arg.equals("-noensemble"))
                mIncludeEnsemble = false;
            else if (arg.equals("-experimentpath"))
                mExperimentPath = argQueue.poll();
            else
                processArg(arg, argQueue);
        }

        //Create an instance of the instance generator
        mInstanceGenerator = InstanceGenerator.create(mExperiment.instanceGenerator, mExperiment.datasetString);

        //Load up all the attribute selectors that we can 
        if(mExperiment.attributeSelection)
            loadAttributeSelectors();
        
        mAllowedClassifiers = exp.allowedClassifiers;

        //Load up all the classifiers for the dataset we can
        loadClassifiers();

        if(mAllowedClassifiers.isEmpty()){
            if(mIncludeBase){
                for(ClassParams clsParams: mBaseClassParams){
                    mAllowedClassifiers.add(clsParams.getTargetClass());
                }
            }
            if(mIncludeMeta){
                for(ClassParams clsParams: mMetaClassParams){
                    mAllowedClassifiers.add(clsParams.getTargetClass());
                }
            }
            if(mIncludeEnsemble){
                for(ClassParams clsParams: mEnsembleClassParams){
                    mAllowedClassifiers.add(clsParams.getTargetClass());
                }
            }
        }

        //Make sure we're conflict free
        checkPrefixes();

        //Make sure that the folder for this experiment exists
        Util.makePath(mExperimentPath + File.separator + mExperiment.name);

        //Generate all the stuff that needs to be created alongside the experiment file
        prepareExperiment(mExperimentPath + File.separator + mExperiment.name + File.separator);

        //Populate the experiment object
        mExperiment.type = getType();
        mExperiment.trajectoryParserClassName = getTrajectoryParserClassName();
        mExperiment.callString = getCallString();
        mExperiment.envVariables = getEnvVariables();
        mExperiment.toXML(mExperimentPath + File.separator + mExperiment.name + File.separator + mExperiment.name + ".experiment");
    }

    /**
     * Useful for subclasses so that they can get some of the handy properties that they'll need to pass to their wrappers
     */
    protected String getWrapperPropString()
    {
        Properties props = new Properties();
        props.setProperty("datasetString", mExperiment.datasetString);
        props.setProperty("instanceGenerator", mExperiment.instanceGenerator);
        props.setProperty("resultMetric", mExperiment.resultMetric);

        /*
        if(mExperiment.regularizer != null)
        {
            sb.append(":regularizer=");
            sb.append(mExperiment.regularizer);

            sb.append(":regularizerParams=");
            if(mExperiment.regularizerArgs != null)
                sb.append(mExperiment.regularizerArgs);
        }*/

        return Util.propertiesToString(props);
    }

    /**
     * Subclasses must provide this method which is responsible for 
     */
    public abstract void prepareExperiment(String path);

    /**
     * Process a constructor argument, and suck out stuff from the arg queue
     */
    public void processArg(String arg, Queue<String> args){
        //Don't do anything by default
    }

    /**
     * Get a string indicating the type of this Experiment (namely the name of the SMBO method)
     */
    protected abstract String getType();
    /**
     * Gets the name of the class that is used to parse the results of the SMBO method into a Trajectory
     */
    protected abstract String getTrajectoryParserClassName();
    /**
     * Gets the set of strings that are called on the command line to invoke the SMBO method
     */
    protected abstract List<String> getCallString();

    /**
     * Gets a list of all the environment variables that need to be set for this experiment to run
     */
    protected List<String> getEnvVariables()
    {
        return Collections.emptyList();
    }
    
    private void loadAttributeSelectors()
    {
        //TODO: Make this a command line option so you can specify bunches of directories
        //TODO: Smartly deal with the fact that maybe these folder doesn't exist
        String[] paramDirs = new String[]{mParamBaseDir};
        Instances instances = mInstanceGenerator.getTraining();

        for(String baseDir: paramDirs)
        {
            //First, process the evaluation methods
            String dir = baseDir + File.separator + "attribselection" + File.separator + "eval" + File.separator;
            ArrayList<String> attributeEvalNames = ApplicabilityTester.getApplicableAttributeEvaluators(instances, dir);

            for(String name:attributeEvalNames)
            {
                ClassParams params = new ClassParams(dir + name + ".params");
                mAttribEvalClassParams.add(params);
            }

            //Next, grab all the search methods
            dir = baseDir + File.separator + "attribselection" + File.separator + "search" + File.separator;
            ArrayList<String> attributeSearchNames = ApplicabilityTester.getApplicableAttributeSearchers(instances, dir);

            for(String name:attributeSearchNames)
            {
                ClassParams params = new ClassParams(dir + name + ".params");
                mAttribSearchClassParams.add(params);
            }
            //if(mAttribEvalClassParams.isEmpty())
                //throw new RuntimeException("Couldn't find any attribute evaluators");
        }
    }

    private void loadClassifiers()
    {
        //TODO: Make this a command line option so you can specify bunches of directories
        //TODO: Smartly deal with the fact that maybe a base/meta/ensemble folder doesn't exist
        String[] paramDirs = new String[]{mParamBaseDir };
        Instances instances = mInstanceGenerator.getTraining();
        List<String> allowed = null;
        if(mAllowedClassifiers.size() > 0)
            allowed = mAllowedClassifiers;

        for(String baseDir: paramDirs)
        {
            //First, process the base classifiers
            String dir = baseDir + File.separator + "base" + File.separator;
            ArrayList<String> baseClassifierNames = ApplicabilityTester.getApplicableClassifiers(instances, dir, allowed);

            for(String name:baseClassifierNames)
            {
                ClassParams params = new ClassParams(dir + name + ".params");
                mBaseClassParams.add(params);
            }

            //Next, process the meta classifiers
            dir = baseDir + File.separator + "meta" + File.separator;
            ArrayList<String> metaClassifierNames = ApplicabilityTester.getApplicableClassifiers(instances, dir, allowed);

            for(String name:metaClassifierNames)
            {
                ClassParams params = new ClassParams(dir + name + ".params");
                mMetaClassParams.add(params);
            }

            //And do all the ensemble methods
            dir = baseDir + File.separator + "ensemble" + File.separator;
            ArrayList<String> ensembleClassifierNames = ApplicabilityTester.getApplicableClassifiers(instances, dir, allowed);

            for(String name:ensembleClassifierNames)
            {
                ClassParams params = new ClassParams(dir+ name + ".params");
                mEnsembleClassParams.add(params);
            }
        }
    }

    private void checkPrefixes()
    {
        ArrayList<String> prefixes = new ArrayList<String>();

        ArrayList<List<ClassParams>> classParams = new ArrayList<List<ClassParams>>();
        classParams.add(mBaseClassParams);
        classParams.add(mMetaClassParams);
        classParams.add(mEnsembleClassParams);
        classParams.add(mAttribEvalClassParams);
        classParams.add(mAttribSearchClassParams);

        for(List<ClassParams> params: classParams){
            for(ClassParams param: params)
            {
                String prefix = getPrefix(param.getTargetClass());
                if(prefixes.contains(prefix))
                {
                    throw new RuntimeException("Prefix '" + prefix + "' (" + param.getTargetClass() + ") is already in used by somebody");
                }
                prefixes.add(prefix);
            }
        }
    }

    /**
     * Gets a prefix out of the classifier name by stripping all the packages and capital letters - needed to ensure that parameters with the same WEKA name don't collide.
     */
    public String getPrefix(String classifierName)
    {
        String prefix = "";
        boolean readDot = true;
        for(int i = 0; i < classifierName.length(); i++)
        {
            char currentChar = classifierName.charAt(i);
            if(readDot || currentChar >= 'A' && currentChar <= 'Z')
            {
                readDot = false;
                prefix += currentChar;
            }
            if(currentChar == '.')
                readDot = true;
        }
        return prefix.toLowerCase();
    }
    
    
    /**
     * Populates a ParameterConditionalGroup with all the params/conditionals that are needed for 
     * optimization methods that support a DAG structure
     */
    public ParameterConditionalGroup generateAlgorithmParameterConditionalGroupForDAG()
    {
        ParameterConditionalGroup paramGroup = new ParameterConditionalGroup();
        //Add in all our our attribute search methods
        if(mExperiment.attributeSelection)
        {
            if(mAttribEvalClassParams.isEmpty()){
                throw new RuntimeException("Couldn't find any attribute eval methods");
            }
            if(mAttribSearchClassParams.isEmpty()){
                throw new RuntimeException("Couldn't find any attribute search methods");
            }

            List<String> searchParamNames = new ArrayList<String>(); 
            //Insert NONE for now, we need to take it out in a second.... after we build the param
            searchParamNames.add("NONE");
            for(ClassParams clsParams: mAttribSearchClassParams){
                searchParamNames.add(clsParams.getTargetClass());
            }
            Parameter attributesearch = new Parameter("attributesearch", searchParamNames);
            paramGroup.add(attributesearch);
            //Take out the NONE
            searchParamNames.remove(0);
            for(ClassParams clsParams: mAttribSearchClassParams){
                addClassifierToParameterConditionalGroupForDAG(paramGroup, clsParams, "assearch_", attributesearch);
            }
            
            //Now we have to build the attribute eval param
            List<String> evalParamNames = new ArrayList<String>(); 
            for(ClassParams clsParams: mAttribEvalClassParams){
                evalParamNames.add(clsParams.getTargetClass());
            }
            Parameter attributeeval = new Parameter("attributeeval", evalParamNames);
            paramGroup.add(attributeeval);
            for(ClassParams clsParams: mAttribEvalClassParams){
                addClassifierToParameterConditionalGroupForDAG(paramGroup, clsParams, "aseval_", attributeeval);
            }

            paramGroup.add(new Parameter("attributetime", "" + mExperiment.attributeSelectionTimeout));
            paramGroup.add(new Conditional(attributeeval, attributesearch, searchParamNames));
        }

        //First, insert the top level choice about which classifer is going to be selected
        List<String> classifiers = new ArrayList<String>();
        List<String> baseClassifiers = new ArrayList<String>();
        List<String> metaClassifiers = new ArrayList<String>();
        List<String> ensembleClassifiers = new ArrayList<String>();
        //Go build up the names of base methods
        for(ClassParams clsParams: mBaseClassParams) { 
            String className = clsParams.getTargetClass();
            classifiers.add(className); 
            baseClassifiers.add(className);
        }
        //Go build up the names of meta methods
        if(mIncludeMeta){
            for(ClassParams clsParams: mMetaClassParams) { 
                String className = clsParams.getTargetClass();
                classifiers.add(className); 
                metaClassifiers.add(className);
            }
        }
        //Go build up the names of ensemble methods
        if(mIncludeEnsemble)
        {
            for(ClassParams clsParams: mEnsembleClassParams) { 
                String className = clsParams.getTargetClass();
                classifiers.add(className); 
                ensembleClassifiers.add(className);
            }
        }

        //Sanity check - we do have one normal classifier?
        if(baseClassifiers.isEmpty())
        {
            throw new RuntimeException("No Base classifiers could be applied to this data set");
        }

        //Build the entire list of all classifiers as a parameter, and insert it
        Parameter targetclass = new Parameter("targetclass", classifiers);
        paramGroup.add(targetclass);

        //Next, insert all the default parameters for each method (Just the flat level
        if(mIncludeBase){
            for(ClassParams clsParams: mBaseClassParams) { 
                addClassifierToParameterConditionalGroupForDAG(paramGroup, clsParams, "_0_", targetclass); 
            }
        }
        if(mIncludeMeta){
            for(ClassParams clsParams: mMetaClassParams) { 
                addClassifierToParameterConditionalGroupForDAG(paramGroup, clsParams, "_0_", targetclass); 
            }
        }
        if(mIncludeEnsemble){
            for(ClassParams clsParams: mEnsembleClassParams) {
                addClassifierToParameterConditionalGroupForDAG(paramGroup, clsParams, "_0_", targetclass); 
            }
        }

        //Add in the meta conditionals
        if(!metaClassifiers.isEmpty())
        {
            Parameter _1_W = new Parameter("_1_W", baseClassifiers);
            paramGroup.add(_1_W);
            paramGroup.add(new Conditional(_1_W, targetclass, metaClassifiers));
            Parameter _1_W_0_DASHDASH = new Parameter("_1_W_0_DASHDASH", "REMOVED");
            paramGroup.add(_1_W_0_DASHDASH);
            paramGroup.add(new Conditional(_1_W_0_DASHDASH, targetclass, metaClassifiers));
            for(ClassParams clsParams: mBaseClassParams) {
                addClassifierToParameterConditionalGroupForDAG(paramGroup, clsParams, "_1_W_1_", _1_W); 
            }
        }

        //Add in the ensemble conditionals
        if(!ensembleClassifiers.isEmpty())
        {
            //Make the ensemble depth param
            List<String> levels = new ArrayList<String>();
            for(int l = 0; l < mEnsembleMaxNum; l++) {
                levels.add(String.format("%d", l));
            }
            Parameter _HIDDEN_ensemble_depth = new Parameter("_HIDDEN_ensemble_depth", levels);
            paramGroup.add(_HIDDEN_ensemble_depth);
            paramGroup.add(new Conditional(_HIDDEN_ensemble_depth, targetclass, ensembleClassifiers));

            //Now actually insert all the levels
            for(int i = 0; i < mEnsembleMaxNum; i++)
            {
                String prefix = "_1_" + String.format("%02d", i);
                Parameter gateParam = new Parameter(prefix + "_0_QUOTE_START_B", baseClassifiers);
                paramGroup.add(gateParam);
                for(ClassParams clsParams: mBaseClassParams) {
                    addClassifierToParameterConditionalGroupForDAG(paramGroup, clsParams, prefix + "_1_", gateParam); 
                }
                
                //Add this level for everything beneath us
                levels.clear();
                for(int l = i; l < mEnsembleMaxNum; l++) {
                    levels.add(String.format("%d", l));
                }
                paramGroup.add(new Conditional(gateParam, _HIDDEN_ensemble_depth, levels));
                //Make the dummy param to close the quote
                Parameter endQuote = new Parameter("_1_" + String.format("%02d", i) + "_2_QUOTE_END", "REMOVED");
                paramGroup.add(endQuote);
                paramGroup.add(new Conditional(endQuote, _HIDDEN_ensemble_depth, levels));
            }
        }

        return paramGroup;
    }

    /** Internal helper method that adds a child classifier's params assuming a DAG structure */
    private void addClassifierToParameterConditionalGroupForDAG(ParameterConditionalGroup paramGroup, ClassParams clsParams, String prefix, Parameter parent)
    {
        Map<Parameter, Parameter> paramMap = new HashMap<Parameter, Parameter>();

        prefix = prefix + "_" + getPrefix(clsParams.getTargetClass()) + "_";

        int i = 0;
        for(Parameter oldParam : clsParams.getParameters())
        {
            //Add in these parameters - but update a map so we can do the conditionals properly in a sec...
            String tempPrefix = prefix + String.format("%02d", i++) + "_";
            Parameter param = new Parameter(tempPrefix + oldParam.name, oldParam);
            paramMap.put(oldParam, param);
            paramGroup.add(param);

            paramGroup.add(new Conditional(param, parent, clsParams.getTargetClass()));
        }
        for(Conditional cond: clsParams.getConditionals())
        {
            paramGroup.add(new Conditional(paramMap.get(cond.parameter), paramMap.get(cond.parent), cond));
        }
    }
};
